/*
 * Copyright (c) 2004 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: flip.c,v 1.2 2004/05/10 22:41:47 zinnia Exp $
 */
#include "sps.h"
//#include "flip.h"

extern SPS_Flip_Vector fader_vector;
extern SPS_Flip_Vector roller_vector;
extern SPS_Flip_Vector iidx7th_vector;
extern SPS_Flip_Vector typoh_vector;
extern SPS_Flip_Vector palette_vector;
extern SPS_Flip_Vector blur_vector;

static SPS_Flip_Vector* flip_drivers[] = {
	&blur_vector,
	&typoh_vector,
	&palette_vector, 
	&iidx7th_vector,
	&roller_vector, // XXX
	&fader_vector,
	NULL
};

static GHashTable* flip_drivers_hash = NULL;

static void implicit_init()
{
	SPS_Flip_Vector** p;
	if(flip_drivers_hash != NULL) return;
	flip_drivers_hash = g_hash_table_new(g_str_hash, g_str_equal);
	for(p = flip_drivers; *p != NULL; p++){
		gpointer oldvalue, oldkey;
		if(g_hash_table_lookup_extended(flip_drivers_hash, 
						(*p)->name, &oldkey, &oldvalue)){
			g_hash_table_remove(flip_drivers_hash,
					    (*p)->name);
		}
		g_hash_table_insert(flip_drivers_hash, (*p)->name, *p);
	}
}

SPS_Flip_Vector* SPS_Flip_Driver_Lookup(const char* name)
{
	SPS_Flip_Vector* result;
	implicit_init();
	
	g_debug("lookup: %s", name);
	result = g_hash_table_lookup(flip_drivers_hash, name);

	if(result == NULL) {
		g_message("%s: lookup failed. using default driver.", name);
		result = SPS_Flip_Driver_Default();
	}
	return result;
}

SPS_Flip_Vector* SPS_Flip_Driver_Default()
{
	return flip_drivers[0];
}
