/*
 * Copyright (c) 2004 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: function.c,v 1.16 2004/04/30 23:42:12 zinnia Exp $
 */

#include "sps.h"

static SPS_Sprite* load_image(const char* path)
{
	SPS_Sprite* result;
	SDL_Surface* surface;
	surface = IMG_Load(path);
	if(surface == NULL) return NULL;

	result = SPS_Sprite_new();
	SPS_Sprite_Set_Surface(result, SDL_ConvertSurface(surface, display->format, SDL_SWSURFACE));

	return result;
}

/* ---------------------------------------- */

void SPS_Fill(SPS_Render_Context* context)
{
// XXXX
}

void SPS_Pause(SPS_Render_Context* context, int timer)
{
	context->pauselevel++;
}

void SPS_MoveCursor(SPS_Render_Context* context, int x, int y)
{
	context->cursor.x = RatioToPixelH(context, x);
	context->cursor.y = RatioToPixelV(context, y);
}


static SPS_DrawUnit* SPS_DrawUnit_new(SPS_Render_Context* context, SPS_DRAWUNIT_TYPE type)
{
	SPS_DrawUnit* u;
	u = g_malloc0(sizeof(SPS_DrawUnit));
	u->pauselevel = context->pauselevel;
	u->lineno = context->lineno;
	u->type = type;
#ifdef USE_OPENGL
	u->sprite.texture = 0;
#endif
	return u;
}

void SPS_DefFont(SPS_Render_Context* context, const char* name, const char* path)
{
	gpointer oldkey, oldvalue;

	if(g_hash_table_lookup_extended(context->fontpath, name, &oldkey, &oldvalue)){
		/* replace */
		g_hash_table_remove(context->fontpath, name);
		g_free(oldkey);
		g_free(oldvalue);
	} 
	g_hash_table_insert(context->fontpath, g_strdup(name), g_strdup(path));
	fprintf(stderr, "fontpath(%s) = %s\n", name, path);
}

void SPS_DefFontIndex(SPS_Render_Context* context, const char* id, const char* name, int point, int aa)
{
	const char* path;
	TTF_Font* font;

	path = g_hash_table_lookup(context->fontpath, name);
	if(path != NULL){
		/* XXX point sanity check */
		font = TTF_OpenFont(path, point);
		if(font != NULL){
			gpointer oldkey, oldvalue;
			if(g_hash_table_lookup_extended(context->fonts, id, &oldkey, &oldvalue)){
				g_hash_table_remove(context->fonts, id);
				g_free(oldkey);
				TTF_CloseFont(oldvalue);
			}
			g_hash_table_insert(context->fonts, g_strdup(id), font);
			fprintf(stderr, "fontindex(%s) = %s at %d point\n", id, path, point);
		} else {
			fprintf(stderr, "TTF_OpenFont: %s\n", SDL_GetError());
		}
	} else {
		fprintf(stderr, "font name '%s' not found.\n", name);
	}
}

static SPS_DrawUnit* SPS_DrawUnit_FromChar(SPS_Render_Context* context, TTF_Font* font, int style, SDL_Color color, gchar* ch)
{
	SPS_DrawUnit* u = NULL; // result
	SPS_Sprite* sprite;

//	g_debug("length = %d", g_utf8_strlen(ch, -1));
	sprite = SPS_Sprite_from_String(font, style, color, ch);
	if(sprite != NULL){
		u = SPS_DrawUnit_new(context, SPS_DRAWUNIT_TYPE_LETTER);
		u->sprite = *sprite;
        u->ch = g_strdup(ch);
	} else {
		fprintf(stderr, "TTF_RenderUTF8_Blend: %s\n", SDL_GetError());
	}
	return u;
}
void SPS_Page_Add_DrawUnit(SPS_Render_Context* context, SPS_DrawUnit* u)
{
	if(context->in_linerender == 0){
		SPS_Page* page = context->page_current;
		if(page != NULL){
			page->drawunits = 
				g_list_append(
					page->drawunits,
					u);
			u->sprite.x = context->cursor.x;
			u->sprite.y = context->cursor.y;
			u->lineno = context->lineno;
			context->cursor.x += u->sprite.w;
		}
	} else {
		context->line = g_list_append(context->line, u);
	}

}

void SPS_PutText(SPS_Render_Context* context, const char* text, const char* fontindex, SPS_VERTICAL_ALIGN align)
{
	SPS_Page* page = context->page_current;
	SDL_Color c;
	TTF_Font* font;
	int i;
	gchar* utf8text;
	font = GetFont(context, fontindex);
	if(font == NULL) {
		fprintf(stderr, "GetFont(%s) == NULL\n", fontindex);
		return;
	}
	SDL_UnMapRGB(display->format, &c, page->setting.forecolor);
	utf8text = get_utf8_string(text);

	for(i = 0; i < g_utf8_strlen(utf8text, -1); i++){
		gchar chp[10]; // XXX ???
		gchar* ch = g_utf8_offset_to_pointer(utf8text, i);
		memset(chp, 0, 10);

		g_utf8_strncpy(chp, ch, 1);

		SPS_DrawUnit* u = 
			SPS_DrawUnit_FromChar(
				context,
				font,
				TTF_GetFontStyle(font),
				c,
				chp);
		if(u == NULL) continue;
		u->valign = align; // XXX messy
		SPS_Page_Add_DrawUnit(context, u);
	}
}


void SPS_ForeColor(SPS_Render_Context* context, const char* color)
{
	Uint32 c = ParseColor(context, color);
	if(context->in_pagerender){
		context->page_current->setting.forecolor = c;
	} else {
		context->setting.forecolor = c;
	}
}

void SPS_BackColor(SPS_Render_Context* context, const char* color)
{
	Uint32 c = ParseColor(context, color);
	if(context->in_pagerender){
		context->page_current->setting.backcolor = c;
	} else {
		context->setting.backcolor = c;
	}
}
void SPS_PutHBar(SPS_Render_Context* context, double thick)
{
	int height;
	SDL_Rect r;
	SPS_Page* page = context->page_current;
	if(page == NULL) return;
	SPS_DrawUnit* u = SPS_DrawUnit_new(context, SPS_DRAWUNIT_TYPE_PRIMITIVE);

	height = RatioToPixelV(context, thick);
	r.w = display->w - context->cursor.x;
	r.h = height;

	SPS_Sprite_Set_Surface(&u->sprite,
			       SDL_CreateRGBSurface(SDL_SWSURFACE,
						    r.w, r.h,
						    display->format->BitsPerPixel,
						    display->format->Rmask,
						    display->format->Gmask,
						    display->format->Bmask,
						    display->format->Amask));
	r.x = r.y = 0;

//	fprintf(stderr, "puthbar: %f -> (%d, %d - %d x %d)\n", thick, r.x, r.y, r.w, r.h);
	SDL_FillRect(u->sprite.surface, &r, page->setting.forecolor);
	u->sprite.w = r.w;
	u->sprite.h = r.h;


	u->sprite.x = page->setting.hmargin;
	u->sprite.y = context->cursor.y;


	page->drawunits = g_list_append(page->drawunits, u);

	context->cursor.x = page->setting.hmargin;
	context->cursor.y += height;
}

void SPS_MoveTo(SPS_Render_Context* context, int x, int y)
{
	int width, height;
	width = RatioToPixelH(context, x);
	height = RatioToPixelV(context, y);
//	fprintf(stderr, "moveto: %d, %d -> %d, %d\n", x, y, width, height);
	context->cursor.x += width;
	context->cursor.y += height;
}

int GetTopLeftY(SPS_VERTICAL_ALIGN valign, int h, int maxheight)
{
	switch(valign){
	case SPS_VERTICAL_ALIGN_TOP:
		return 0;
	case SPS_VERTICAL_ALIGN_CENTER:
		return (maxheight - h) / 2;
	case SPS_VERTICAL_ALIGN_BOTTOM:
		return maxheight - h;
	}
	return 0;
}

int GetTopLeftX(SPS_HORIZONTAL_ALIGN halign, int w, int maxwidth)
{
	switch(halign)
	{
	case SPS_HORIZONTAL_ALIGN_LEFT:
		return 0;
	case SPS_HORIZONTAL_ALIGN_CENTER:
		return (maxwidth - w) / 2;
	case SPS_HORIZONTAL_ALIGN_RIGHT:
		return maxwidth - w;
	}
	return 0;
}

static void render_1line(SPS_Render_Context* context, GList* line, int maxh)
{
	int maxw = display->w;
	int totalw = 0;
	int x;
	GList* p;
	SPS_Page* page;
	SPS_DrawUnit* u;

	page = context->page_current;
	if(page == NULL) return;
	x = context->cursor.x;

//	fprintf(stderr, "render_1line: %d\n", g_list_length(line));
	// Phase1: determine total width and vertical location
	G_LIST_FOREACH(line, p){
		u = p->data;
		u->sprite.y = context->cursor.y + GetTopLeftY(u->valign, u->sprite.h, maxh);
		totalw += u->sprite.w;
	}
	// horizontal offset
	x += GetTopLeftX(page->setting.halign, totalw, maxw - context->cursor.x);

	// Phase2: determine horizontal location and register it
	G_LIST_FOREACH(line, p){
		u = p->data;
		u->sprite.x = x;
		x += u->sprite.w;

		page->drawunits = 
			g_list_append(
				page->drawunits,
				u);
	}
}

void SPS_EndLine(SPS_Render_Context* context)
{
	SPS_Page* page = context->page_current;
	GList* p;
	int maxh;
	int maxw = display->w;
	GList* linetmp = NULL;

	maxh = 0;
	int x;

	x = context->cursor.x;
//	fprintf(stderr, "SPS_EndLine: %d\n", g_list_length(context->line));
	// SPS_Text_Chunk$B$r$=$l$>$lI>2A(B
	G_LIST_FOREACH(context->line, p){
		SPS_DrawUnit* u = p->data;

		if(x + u->sprite.w >= maxw){
			render_1line(context, linetmp, maxh);
			linetmp = NULL; // XXX

			x = 0;
			// next line
			context->cursor.x = page->setting.hmargin;
			context->cursor.y += maxh;
			context->lineno ++;
			maxh = 0;
		} 
		linetmp = g_list_append(linetmp, u);
		x += u->sprite.w;

		if(maxh < u->sprite.h) maxh = u->sprite.h;
		u->lineno = context->lineno;
	}
	render_1line(context, linetmp, maxh);
	context->cursor.x = page->setting.hmargin; // XXX?
	context->cursor.y += maxh;
	context->lineno++;
}

void SPS_BeginLine(SPS_Render_Context* context, const char* halign)
{
	SPS_Page* page = context->page_current;
	GList* p;
	context->in_linerender = 1;

	page->setting.halign = SPS_Horizontal_Align_Parse(halign);
//	fprintf(stderr, "beginline\n");
	context->cursor.x = page->setting.hmargin;

	if(context->line != NULL){
		G_LIST_FOREACH(context->line, p){

		}
		g_list_free(context->line);
	}
	context->line = NULL;
}


void SPS_FontStyle(SPS_Render_Context* context, const char* fontindex, const char* expression)
{
	int style;
	TTF_Font* font = GetFont(context, fontindex);
	if(font == NULL) return;
	style = TTF_Style_Parse(expression);
	TTF_SetFontStyle(font, style);
	g_debug("TTF_SetFontStyle: %d", style);
}

void SPS_HMargin(SPS_Render_Context* context, int m)
{
	int l = RatioToPixelH(context, m);
	if(context->in_pagerender){
		context->page_current->setting.hmargin = l;
	} else {
		context->setting.hmargin = l;
	}
}

void SPS_DefIcon(SPS_Render_Context* context, const char* no, const char* filename)
{
	gpointer oldvalue, oldkey;
	SPS_Sprite* sprite = load_image(filename);
	if(sprite == NULL) return;

	if(g_hash_table_lookup_extended(context->icons, no, &oldkey, &oldvalue)){
		g_hash_table_remove(context->icons, no);
		g_free(oldkey);
		SPS_Sprite_free(oldvalue);
	}
	g_hash_table_insert(context->icons, g_strdup(no), sprite);
//	fprintf(stderr, "icon(%s) = %s\n", no, filename);
}

void SPS_PutIcon(SPS_Render_Context* context, const char* iconindex, SPS_VERTICAL_ALIGN align, int width, int height)
{
	SPS_Sprite* sprite = GetIcon(context, iconindex);
	if(sprite == NULL) return;
	SPS_DrawUnit* u = SPS_DrawUnit_new(context, SPS_DRAWUNIT_TYPE_PRIMITIVE);
	u->sprite = *sprite;
	if(width >= 0) u->sprite.w = width; 
	else u->sprite.w = u->sprite.surface->w;
	if(height >= 0) u->sprite.h = height;
	else u->sprite.h = u->sprite.surface->h;

	u->valign = align;
	SPS_Page_Add_DrawUnit(context, u);
}

void SPS_PutImage(SPS_Render_Context* context, const char* path, SPS_VERTICAL_ALIGN align, int width, int height)
{
	SPS_Page* page = context->page_current;
	SPS_Sprite* sprite = load_image(path);
	if(sprite == NULL) return;

	SPS_DrawUnit* u = 
		SPS_DrawUnit_new(context, SPS_DRAWUNIT_TYPE_PRIMITIVE);
	u->sprite = *sprite;
	if(width >= 0) u->sprite.w = width;
	if(height >= 0) u->sprite.h = height;

	u->valign = align;
	SPS_Page_Add_DrawUnit(context, u);
}
			       
void SPS_Bg(SPS_Render_Context* context, const char* path,
	    SPS_BGSTYLE style, int alpha)
{
	SPS_Page* page = context->page_current;

	SPS_Sprite* sprite = load_image(path);

	if(context->in_pagerender){
		page->setting.bg_style = style;
		page->setting.bg_alpha = (double)alpha / 100.0;
		page->setting.bg_sprite = sprite;
	} else {
		context->setting.bg_style = style;
		context->setting.bg_alpha = (double)alpha / 100.0;
		context->setting.bg_sprite = sprite;
	}
}

void SPS_BeginPage(SPS_Render_Context* context, char* title)
{
//	fprintf(stderr, "beingpage()\n");
	context->cursor.x = 0;
	context->cursor.y = 0;
	context->setting.hmargin = 0;

	context->pauselevel = 0;
	context->lineno = 0;

	// XXX backcolor/forecolor/halign/hmargin
	context->in_linerender = 0;
	context->in_pagerender = 1;

	SPS_Page* page = context->page_current;
	page->setting = context->setting;
	page->title = g_strdup(title);
}

void SPS_EndPage(SPS_Render_Context* context)
{
	SPS_Page* page = context->page_current;
//	fprintf(stderr, "endpage\n");
	page->pauselevel_max = context->pauselevel;
	context->in_pagerender = 0;
}

void SPS_Flip_Function(SPS_Render_Context* context, const char* name)
{
// XXX $B<B9T;~%3%s%F%-%9%H$J$s$@$h$M$((B
// $B$H$j$"$($:;~4V$b$J$$$N$G<jH4$-<BAu(B(2004/4/30)
	SPS_Page* page = context->page_current;
	SPS_Flip_Vector* v = SPS_Flip_Driver_Lookup(name); 
	if(context->in_pagerender){
		page->setting.flip_driver = v;
	} else {
		context->setting.flip_driver = v;
	}
}

void SPS_SetTitle(SPS_Render_Context* context, const char* name)
{
	context->presentation->title = g_strdup(name);
}
