/*
 * Copyright (c) 2004 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: sps_sprite.c,v 1.5 2004/05/18 03:49:36 zinnia Exp $
 */

#include "sps.h"

SPS_Sprite* SPS_Sprite_new()
{
	SPS_Sprite* result = (SPS_Sprite*)g_malloc0(sizeof(SPS_Sprite));
#ifdef USE_OPENGL
	result->texture = 0;
#endif
	result->surface = NULL;
	result->surface_dirty = FALSE;
	result->x = result->y = 0;
	result->userdata = NULL;
}

void SPS_Sprite_Free_Surface(SPS_Sprite* sprite)
{
	if(sprite->surface != NULL){
		SDL_FreeSurface(sprite->surface);
		sprite->surface = NULL;
		sprite->surface_dirty = FALSE;
	}
}

void SPS_Sprite_Free_Texture(SPS_Sprite* sprite)
{
#ifdef USE_OPENGL
	if(display->use_opengl){
		if(sprite->texture != 0){
			glDeleteTextures(1, &(sprite->texture));
			sprite->texture = 0;
		}
	} else 
#endif
	{
	}
}
SPS_Sprite* SPS_Sprite_free(SPS_Sprite* sprite)
{
	SPS_Sprite_Free_Texture(sprite);
	SPS_Sprite_Free_Surface(sprite);
	if(sprite->userdata_autofree && sprite->userdata != NULL){
		g_free(sprite->userdata);
		sprite->userdata = NULL;
	}
	free(sprite);
	return NULL;
}

void SPS_Sprite_Set_Surface(SPS_Sprite* sprite, SDL_Surface* surface)
{
	if(surface != NULL){
		sprite->surface = surface;
		sprite->surface_dirty = TRUE;
		sprite->w = surface->w;
		sprite->h = surface->h;
	}
}
SPS_Sprite* SPS_Sprite_new_WithSize(int width, int height)
{
	SPS_Sprite* result = SPS_Sprite_new();
	SPS_Sprite_Set_Surface(result,
			       SDL_CreateRGBSurface(SDL_SWSURFACE,
						    width, height,
						    display->format->BitsPerPixel,
						    display->format->Rmask,
						    display->format->Gmask,
						    display->format->Bmask,
						    display->format->Amask));
			       return result;
}

SPS_Sprite* SPS_Sprite_from_String(TTF_Font* font, int style, SDL_Color color, gchar* string)
{
	SPS_Sprite* result = NULL;
	SDL_Surface* rendered;

	if(font == NULL) return;
	if(string == NULL) return;

	TTF_SetFontStyle(font, style);
	rendered = TTF_RenderUTF8_Blended(font, string, color);

	if(rendered != NULL){
		result = SPS_Sprite_new();
		SPS_Sprite_Set_Surface(result, rendered);
	} else {
		fprintf(stderr, "SPS_Sprite_from_String: %s", SDL_GetError());
	}
	return result;
}

int power_of_two(int input)
{
	int value = 1;

	while ( value < input ) {
		value <<= 1;
	}
	return value;
}
#ifdef USE_OPENGL
static void LoadTexture(SPS_Sprite* sprite, gboolean subimage)
{
	int i, j, bpp;
	int w, h;
	SDL_Surface* surface = sprite->surface;

	w = power_of_two(surface->w);
	h = power_of_two(surface->h);

	// 2^nˤ碌ƤΤǡ;ʬʤȤˤϥߤäƤޤ

	sprite->texcoord_minx = 0.0f;			/* Min X */
	sprite->texcoord_miny = 0.0f;			/* Min Y */
	sprite->texcoord_maxx = (GLfloat)surface->w / (GLfloat)w;	/* Max X */
	sprite->texcoord_maxy = (GLfloat)surface->h / (GLfloat)h;	/* Max Y */

	bpp = surface->format->BitsPerPixel;
	if(bpp != 24 && bpp != 32)
	{
		return;
	}

	//Uint8* pbuf = (Uint8*)malloc(w * h * bpp / 8);
	Uint8* pbuf = (Uint8*)g_malloc(w * h * bpp / 8);
	Uint8* psrc, * pdest;
	for(i = 0;i < surface->h;i++)
	{
		psrc = (Uint8*) surface->pixels + surface->pitch * i;
		pdest = pbuf + i * w * bpp / 8;

		for(j = 0;j < surface->w;j++)
		{
			Uint8 r, g, b, a;
			SDL_GetRGBA(*((Uint32*)psrc), surface->format, &r, &g, &b, &a);
			pdest[0] = r;
			pdest[1] = g;
			pdest[2] = b;

			if(bpp == 32)
			{
				pdest[3] = a;
			}
			psrc += surface->format->BytesPerPixel;
			pdest += bpp / 8;
		}
	}
	Uint32 format;

	if(bpp == 24)
	{
		format = GL_RGB;
	}
	else
	{
		format = GL_RGBA;
	}

	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	if(subimage == FALSE){
		glGenTextures(1,&(sprite->texture));
		glBindTexture(GL_TEXTURE_2D,sprite->texture);
              
		gluBuild2DMipmaps(GL_TEXTURE_2D,
				  format,
				  w,
				  h,
				  format,
				  GL_UNSIGNED_BYTE,
				  pbuf);
	} else {
		glBindTexture(GL_TEXTURE_2D,sprite->texture);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
				w, h,
				GL_RGBA, GL_UNSIGNED_BYTE, pbuf);
	}
	free(pbuf);
}
#endif

void SPS_Sprite_Bind(SPS_Sprite* sprite)
{
#ifdef USE_OPENGL
	if(display->use_opengl){
		if(sprite->texture == 0){
			// bind
			LoadTexture(sprite, FALSE);
			sprite->surface_dirty = FALSE;
		}
		if(sprite->surface_dirty){
			// subimage
			LoadTexture(sprite, TRUE);
			sprite->surface_dirty = FALSE;
		}
		glBindTexture(GL_TEXTURE_2D, sprite->texture);
	} else 
#endif
	{
	}
	display->textures ++;
}

void SPS_Sprite_Draw_Generic(SPS_Sprite* sprite, SPS_DRAW_PARAMETER_ENUM param_enum, SPS_Draw_Parameter* param)
{
	int x, y;
	if(sprite == NULL) return;
	if(param_enum & SPS_DRAW_PARAMETER_ENUM_POSITION){
		x = param->x;
		y = param->y;
	} else {
		x = sprite->x;
		y = sprite->y;
	}

	SPS_Sprite_Bind(sprite);
	if(param->shade &&
	   (param_enum & SPS_DRAW_PARAMETER_ENUM_ALPHA) && 
	   display->sprite_shade &&
#ifdef USE_OPENGL
	   display->use_opengl
#else
	   FALSE
#endif
){
		SPS_Draw_Parameter p;
		p.x = x + 2;
		p.y = y + 2;
		p.shade = FALSE;
		p.alpha = param->alpha * 0.3;
		SPS_Sprite_Draw_Generic(sprite, param_enum, &p);

		p.x += 1;
		p.y += 1;
		p.alpha = param->alpha * 0.2;
		SPS_Sprite_Draw_Generic(sprite, param_enum, &p);

		p.x += 1;
		p.y += 1;
		p.alpha = param->alpha * 0.1;
		SPS_Sprite_Draw_Generic(sprite, param_enum, &p);
	}

#ifdef USE_OPENGL
	if(display->use_opengl){
		if(param_enum & SPS_DRAW_PARAMETER_ENUM_ALPHA){
			glColor4d(1.0, 1.0, 1.0, param->alpha);
		}
		glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2f(sprite->texcoord_minx, sprite->texcoord_miny);
		glVertex2i(x,   y);

		glTexCoord2f(sprite->texcoord_maxx, sprite->texcoord_miny); 
		glVertex2i(x + sprite->w, y);

		glTexCoord2f(sprite->texcoord_minx, sprite->texcoord_maxy); 
		glVertex2i(x,   y + sprite->h);
		
		glTexCoord2f(sprite->texcoord_maxx, sprite->texcoord_maxy);
		glVertex2i(x + sprite->w, y + sprite->h);

		glEnd();
	} else
#endif
	{
		SDL_Rect r;
		r.x = x + display->w / 2;
		r.y = y + display->h / 2;
//		r.h = sprite->w;
//		r.w = sprite->h;
		r.w = sprite->surface->w;
		r.h = sprite->surface->h;

		if(param_enum & SPS_DRAW_PARAMETER_ENUM_ALPHA){
			SDL_SetAlpha(sprite->surface, SDL_SRCALPHA |
				     SDL_RLEACCEL, (Uint8)(param->alpha * 255.0));
		}
		SDL_BlitSurface(sprite->surface, NULL, display->display, &r);
	}
}

